" ToolDelta 常量定义 "

PRG_NAME = "ToolDelta"
"程序名"

PLUGIN_MARKET_SOURCE_OFFICIAL = "https://tdload.tblstudio.cn/https://raw.githubusercontent.com/ToolDelta/ToolDelta-PluginMarket/main"
"插件市场源"

LAUNCH_CFG: dict = {
    "启动器启动模式(请不要手动更改此项, 改为0可重置)": 0,
    "验证服务器地址(更换时记得更改fbtoken)": "",
    "是否记录日志": True,
    "是否使用github镜像": True,
    "插件市场源": PLUGIN_MARKET_SOURCE_OFFICIAL
}
"默认登录配置"

LAUNCH_CFG_STD: dict = {
    "启动器启动模式(请不要手动更改此项, 改为0可重置)": int,
    "验证服务器地址(更换时记得更改fbtoken)": str,
    "是否记录日志": bool,
    "是否使用github镜像": bool,
    "插件市场源": str
}
"默认登录配置标准验证格式"

LAUNCHER_NEOMEGA_STD: dict = {
    "服务器号": int,
    "密码": int,
    "验证服务器地址(更换时记得更改fbtoken)": str
}
"启动器：NeOmega 启动配置验证格式"

LAUNCHER_NEOMEGA_DEFAULT: dict = {
    "服务器号": 0,
    "密码": 0,
    "验证服务器地址(更换时记得更改fbtoken)": ""
}
"启动器：NeOmega 默认启动配置"

LAUNCHER_BEWS_STD: dict = {
    "服务端开放地址": str
}
"启动器：BEWSServer 启动配置验证格式"

LAUNCHER_BEWS_DEFAULT: dict = {
    "服务端开放地址": ""
}
"启动器：BEWSServer 默认启动配置"

FB_APIS = [
    "https://api.fastbuilder.pro/api/phoenix/login",
    "https://api.fastbuilder.pro/api/new",
    "https://api.fastbuilder.pro/api/api",
    "https://api.fastbuilder.pro/api/login",
    "https://api.fastbuilder.pro",
]
"验证服务器：FastBuilder API 列表"


GUGU_APIS = [
    "https://liliya233.uk/api/phoenix/login",
    "https://liliya233.uk/api/new",
    "https://liliya233.uk",
]
"验证服务器：Liliya API 列表"

AUTH_SERVERS = [
    ("FastBuilder 官方验证服务器", "https://api.fastbuilder.pro"),
    ("咕咕酱 FB验证服务器", "https://liliya233.uk"),
]
"验证服务器列表"

TOOLDELTA_PLUGIN_DIR = "插件文件"
"插件文件路径"

TOOLDELTA_CLASSIC_PLUGIN = "ToolDelta类式插件"
"插件文件：ToolDelta 类式插件 路径"

TOOLDELTA_INJECTED_PLUGIN = "ToolDelta注入式插件"
"插件文件：ToolDelta 注入式插件 路径"

TOOLDELTA_PLUGIN_DATA_DIR = "插件数据文件"
"插件数据文件文件夹路径"

PLUGIN_TYPE_MAPPING = {
    "classic": TOOLDELTA_CLASSIC_PLUGIN,
    "injected": TOOLDELTA_INJECTED_PLUGIN
}
"插件属性名映射"

TDSPECIFIC_MIRROR = "https://tdload.tblstudio.cn"
"ToolDelta专用镜像"


"数据包类构建器"

from dataclasses import dataclass

@dataclass
class PacketIDS:
    "数据包 ID 常量表"
    IDLogin = 1  # 登录
    IDPlayStatus = 2  # 播放地址
    IDServerToClientHandshake = 3  # 服务器到客户端握手
    IDClientToServerHandshake = 4  # 客户端到服务器握手
    IDDisconnect = 5  # 断开连接
    IDResourcePacksInfo = 6  # 资源包信息
    IDResourcePackStack = 7  # 资源包堆栈
    IDResourcePackClientResponse = 8  # 资源包客户端响应
    Text = 9  # 文本
    IDSetTime = 10  # 设置时间
    IDStartGame = 11  # 开始游戏
    AddPlayer = 12  # 添加玩家
    IDAddActor = 13  # 添加演员
    IDRemoveActor = 14  # 移除演员
    IDAddItemActor = 15  # 添加物品演员
    IDTakeItemActor = 17  # 拿取物品演员
    IDMoveActorAbsolute = 18  # 绝对移动演员
    IDMovePlayer = 19  # 移动玩家
    IDPassengerJump = 20  # 乘客跳跃
    IDUpdateBlock = 21  # 更新方块
    IDAddPainting = 22  # 添加绘画
    IDTickSync = 23  # 时序同步
    IDLevelEvent = 25  # 关卡事件
    IDBlockEvent = 26  # 方块事件
    IDActorEvent = 27  # 演员事件
    IDMobEffect = 28  # 生物效果
    IDUpdateAttributes = 29  # 更新属性
    IDInventoryTransaction = 30  # 库存交易
    IDMobEquipment = 31  # 生物装备
    IDMobArmourEquipment = 32  # 生物护甲装备
    IDInteract = 33  # 交互
    IDBlockPickRequest = 34  # 方块拾取请求
    IDActorPickRequest = 35  # 演员拾取请求
    IDPlayerAction = 36  # 玩家动作
    IDHurtArmour = 38  # 伤害护甲
    IDSetActorData = 39  # 设置演员数据
    IDSetActorMotion = 40  # 设置演员运动
    IDSetActorLink = 41  # 设置演员链接
    IDSetHealth = 42  # 设置生命值
    IDSetSpawnPosition = 43  # 设置生成位置
    IDAnimate = 44  # 动画
    IDRespawn = 45  # 重生
    IDContainerOpen = 46  # 容器打开
    IDContainerClose = 47  # 容器关闭
    IDPlayerHotBar = 48  # 玩家热键栏
    IDInventoryContent = 49  # 库存内容
    IDInventorySlot = 50  # 库存槽
    IDContainerSetData = 51  # 容器设置数据
    IDCraftingData = 52  # 合成数据
    IDCraftingEvent = 53  # 合成事件
    IDGUIDataPickItem = 54  # GUI 数据选择物品
    IDAdventureSettings = 55  # 冒险设置
    IDBlockActorData = 56  # 方块演员数据
    IDPlayerInput = 57  # 玩家输入
    IDLevelChunk = 58  # 关卡区块
    IDSetCommandsEnabled = 59  # 设置命令启用
    IDSetDifficulty = 60  # 设置难度
    IDChangeDimension = 61  # 改变维度
    IDSetPlayerGameType = 62  # 设置玩家游戏类型
    PlayerList = 63  # 玩家列表
    IDSimpleEvent = 64  # 简单事件
    IDEvent = 65  # 事件
    IDSpawnExperienceOrb = 66  # 生成经验球
    IDClientBoundMapItemData = 67  # 客户端限定地图物品数据
    IDMapInfoRequest = 68  # 地图信息请求
    IDRequestChunkRadius = 69  # 请求区块半径
    IDChunkRadiusUpdated = 70  # 区块半径更新
    IDItemFrameDropItem = 71  # 物品框丢弃物品
    IDGameRulesChanged = 72  # 游戏规则变更
    IDCamera = 73  # 摄像头
    IDBossEvent = 74  # 首领事件
    IDShowCredits = 75  # 展示信用
    IDAvailableCommands = 76  # 可用命令
    IDCommandRequest = 77  # 命令请求
    IDCommandBlockUpdate = 78  # 命令方块更新
    CommandOutput = 79  # 命令输出
    IDUpdateTrade = 80  # 更新交易
    IDUpdateEquip = 81  # 更新装备
    IDResourcePackDataInfo = 82  # 资源包数据信息
    IDResourcePackChunkData = 83  # 资源包块数据
    IDResourcePackChunkRequest = 84  # 资源包块请求
    IDTransfer = 85  # 传输
    IDPlaySound = 86  # 播放声音
    IDStopSound = 87  # 停止声音
    IDSetTitle = 88  # 设置标题
    IDAddBehaviourTree = 89  # 添加行为树
    IDStructureBlockUpdate = 90  # 结构块更新
    IDShowStoreOffer = 91  # 展示商店提供
    IDPurchaseReceipt = 92  # 购买凭据
    IDPlayerSkin = 93  # 玩家皮肤
    IDSubClientLogin = 94  # 子客户端登录
    IDAutomationClientConnect = 95  # 自动化客户端连接
    IDSetLastHurtBy = 96  # 设置最后受伤者
    IDBookEdit = 97  # 编辑书籍
    IDNPCRequest = 98  # NPC 请求
    IDPhotoTransfer = 99  # 照片传输
    IDModalFormRequest = 100  # 模态窗口请求
    IDModalFormResponse = 101  # 模态窗口响应
    IDServerSettingsRequest = 102  # 服务器设置请求
    IDServerSettingsResponse = 103  # 服务器设置响应
    IDShowProfile = 104  # 展示配置文件
    IDSetDefaultGameType = 105  # 设置默认游戏类型
    IDRemoveObjective = 106  # 移除目标
    IDSetDisplayObjective = 107  # 设置显示目标
    IDSetScore = 108  # 设置分数
    IDLabTable = 109  # 实验室桌
    IDUpdateBlockSynced = 110  # 更新同步方块
    IDMoveActorDelta = 111  # 移动演员增量
    IDSetScoreboardIdentity = 112  # 设置计分板标识
    IDSetLocalPlayerAsInitialised = 113  # 设置初始本地玩家
    IDUpdateSoftEnum = 114  # 更新软枚举
    IDNetworkStackLatency = 115  # 网络堆栈延迟
    IDScriptCustomEvent = 117  # 脚本自定义事件
    IDSpawnParticleEffects = 118  # 生成粒子效果
    IDAvailableActorIdentifiers = 119  # 可用演员标识
    IDNetworkChunKPublisherUpdate = 121  # 网络区块发布者更新
    IDBiomeDefinitionList = 122  # 生物群落定义列表
    IDLevelSoundEvent = 123  # 关卡声音事件
    IDLevelEventGeneric = 124  # 关卡通用事件
    IDLecternUpdate = 125  # 讲台更新
    IDAddEntity = 127  # 添加实体
    IDRemoveEntity = 128  # 移除实体
    IDClientCacheStatus = 129  # 客户端缓存状态
    IDMapCreateLockedCopy = 130  # 创建地图锁定副本
    IDOnScreenTextureAnimation = 131  # 屏幕上纹理动画
    IDStructureTemplateDataRequest = 132  # 结构模板数据请求
    IDStructureTemplateDataResponse = 133  # 结构模板数据响应
    IDClientCacheBlobStatus = 135  # 客户端缓存块状态
    IDClientCacheMissResponse = 136  # 客户端缓存未命中响应
    IDEducationSettings = 137  # 教育设置
    IDEmote = 138  # 表情
    IDMultiPlayerSettings = 139  # 多人游戏设置
    IDSettingsCommand = 140  # 设置命令
    IDAnvilDamage = 141  # 铁砧伤害
    IDCompletedUsingItem = 142  # 使用物品完成
    IDNetworkSettings = 143  # 网络设置
    IDPlayerAuthInput = 144  # 玩家认证输入
    IDCreativeContent = 145  # 创造性内容
    IDPlayerEnchantOptions = 146  # 玩家附魔选项
    IDItemStackRequest = 147  # 物品堆请求
    IDItemStackResponse = 148  # 物品堆响应
    IDPlayerArmourDamage = 149  # 玩家护甲伤害
    IDCodeBuilder = 150  # 代码构建
    IDUpdatePlayerGameType = 151  # 更新玩家游戏类型
    IDEmoteList = 152  # 表情列表
    IDPositionTrackingDBServerBroadcast = 153  # 位置追踪数据库服务器广播
    IDPositionTrackingDBClientRequest = 154  # 位置追踪数据库客户端请求
    IDDebugInfo = 155  # 调试信息
    IDPacketViolationWarning = 156  # 数据包违规警告
    IDMotionPredictionHints = 157  # 运动预测提示
    IDAnimateEntity = 158  # 动画实体
    IDCameraShake = 159  # 摄像头震动
    IDPlayerFog = 160  # 玩家迷雾
    IDCorrectPlayerMovePrediction = 161  # 纠正玩家移动预测
    IDItemComponent = 162  # 物品组件
    IDFilterText = 163  # 过滤文本
    IDClientBoundDebugRenderer = 164  # 客户端限定调试渲染器
    IDSyncActorProperty = 165  # 同步演员属性
    IDAddVolumeEntity = 166  # 添加体积实体
    IDRemoveVolumeEntity = 167  # 移除体积实体
    IDSimulationType = 168  # 模拟类型
    IDNPCDialogue = 169  # NPC 对话
    IDEducationResourceURI = 170  # 教育资源 URI
    IDCreatePhoto = 171  # 创建照片
    IDUpdateSubChunkBlocks = 172  # 更新子区块方块
    IDPhotoInfoRequest = 173  # 照片信息请求
    IDSubChunk = 174  # 子区块
    IDSubChunkRequest = 175  # 子区块请求
    IDClientStartItemCooldown = 176  # 客户端开始物品冷却
    IDScriptMessage = 177  # 脚本消息
    IDCodeBuilderSource = 178  # 代码构建源
    IDTickingAreasLoadStatus = 179  # 区域加载状态
    IDDimensionData = 180  # 维度数据
    IDAgentAction = 181  # 代理行动
    IDChangeMobProperty = 182  # 改变生物属性
    IDPyRpc = 183  # Python 远程过程调用

@dataclass
class ActorEventType:
    "演员事件常量表"
    ActorEventJump = 1  # 角色跳跃
    ActorEventHurt = 2  # 角色受伤
    ActorEventDeath = 3  # 角色死亡
    ActorEventStartAttacking = 4  # 开始攻击
    ActorEventStopAttacking = 5  # 停止攻击
    ActorEventTamingFailed = 6  # 驯服失败
    ActorEventTamingSucceeded = 7  # 驯服成功
    ActorEventShakeWetness = 8  # 摇动湿度
    ActorEventUseItem = 9  # 使用物品
    ActorEventEatGrass = 10  # 吃草
    ActorEventFishhookBubble = 11  # 鱼钩泡泡
    ActorEventFishhookFishPosition = 12  # 鱼钩鱼位置
    ActorEventFishhookHookTime = 13  # 鱼钩勾住时间
    ActorEventFishhookTease = 14  # 鱼钩逗弄
    ActorEventSquidFleeing = 15  # 乌贼逃离
    ActorEventZombieConverting = 16  # 僵尸转换
    ActorEventPlayAmbient = 17  # 播放环境音乐
    ActorEventSpawnAlive = 18  # 产生生物
    ActorEventStartOfferFlower = 19  # 开始献花
    ActorEventStopOfferFlower = 20  # 停止献花
    ActorEventLoveHearts = 21  # 爱心
    ActorEventVillagerAngry = 22  # 村民生气
    ActorEventVillagerHappy = 23  # 村民开心
    ActorEventWitchHatMagic = 24  # 女巫帽子魔法
    ActorEventFireworksExplode = 25  # 烟花爆炸
    ActorEventInLoveHearts = 26  # 爱心飞扬
    ActorEventSilverfishMergeAnimation = 27  # 银鱼合并动画
    ActorEventGuardianAttackSound = 28  # 守卫者攻击音效
    ActorEventDrinkPotion = 29  # 喝药水
    ActorEventThrowPotion = 30  # 扔药水
    ActorEventCartWithPrimeTNT = 31  # 载有TNT的车
    ActorEventPrimeCreeper = 32  # 充能的爬行者
    ActorEventAirSupply = 33  # 空气供应
    ActorEventAddPlayerLevels = 34  # 增加玩家等级
    ActorEventGuardianMiningFatigue = 35  # 守卫者挖掘疲劳
    ActorEventAgentSwingArm = 36  # 代理摆动手臂
    ActorEventDragonStartDeathAnim = 37  # 龙开始死亡动画
    ActorEventGroundDust = 38  # 地面灰尘
    ActorEventShake = 39  # 摇动
    ActorEventFeed = 57  # 喂养角色
    ActorEventBabyEat = 60 # 宝宝吃东西
    ActorEventInstantDeath = 61 # 瞬间死亡
    ActorEventNotifyTrade = 62 # 通知交易
    ActorEventLeashDestroyed = 63 # 绳索被销毁
    ActorEventCaravanUpdated = 64 # 更新马车
    ActorEventTalismanActivate = 65 # 法宝激活
    ActorEventUpdateStructureFeature = 66 # 更新结构特征
    ActorEventPlayerSpawnedMob = 67 # 玩家产生的生物
    ActorEventPuke = 68 # 呕吐
    ActorEventUpdateStackSize = 69 # 更新堆叠大小
    ActorEventStartSwimming = 70 # 开始游泳
    ActorEventBalloonPop = 71 # 气球爆破
    ActorEventTreasureHunt = 72 # 寻宝
    ActorEventSummonAgent = 73 # 召唤代理
    ActorEventFinishedChargingCrossbow = 74 # 弩完成充能
    ActorEventLandedOnGround = 75 # 降落在地面
    ActorEventActorGrowUp = 76 # 角色成长
    ActorEventVibrationDetected = 77 # 检测到振动
    
@dataclass
class EventType:
    "事件类型常量表"
    EventTypeAchievementAwarded  = 0  # 成就奖励
    EventTypeEntityInteract  = 1  # 实体交互
    EventTypePortalBuilt  = 2  # 传送门建立
    EventTypePortalUsed  = 3  # 传送门使用
    EventTypeMobKilled  = 4  # 生物被杀
    EventTypeCauldronUsed  = 5  # 釜子使用
    EventTypePlayerDied  = 6  # 玩家死亡
    EventTypeBossKilled  = 7  # Boss被杀
    EventTypeAgentCommand  = 8  # 代理命令
    EventTypeAgentCreated  = 9  # 代理创建（目前无用？）
    EventTypePatternRemoved  = 10  # 模式移除
    EventTypeSlashCommandExecuted  = 11  # 斜杠命令执行
    EventTypeFishBucketed  = 12  # 鱼桶
    EventTypeMobBorn  = 13  # 生物出生
    EventTypePetDied  = 14  # 宠物死亡
    EventTypeCauldronInteract  = 15  # 釜子交互
    EventTypeComposterInteract  = 16  # 堆肥桶交互
    EventTypeBellUsed  = 17  # 钟声
    EventTypeEntityDefinitionTrigger  = 18  # 实体定义触发
    EventTypeRaidUpdate  = 19  # 袭击更新
    EventTypeMovementAnomaly  = 20  # 移动异常
    EventTypeMovementCorrected  = 21  # 移动校正
    EventTypeExtractHoney  = 22  # 提取蜂蜜
    EventTypeTargetBlockHit  = 23  # 目标方块被击中
    EventTypePiglinBarter  = 24  # 畏怯者交易
    EventTypePlayerWaxedOrUnwaxedCopper  = 25  # 玩家给覆蜡/未覆蜡铜
    EventTypeCodeBuilderRuntimeAction  = 26  # 代码构建器运行时动作
    EventTypeCodeBuilderScoreboard  = 27  # 代码构建器计分板
    EventTypeStriderRiddenInLavaInOverworld  = 28  # 熔岩中骑猪灵
    EventTypeSneakCloseToSculkSensor  = 29  # 靠近鸾石感应器潜行

@dataclass
class PlayerActionType:
    PlayerActionStartBreak = 0  # 玩家开始破坏
    PlayerActionAbortBreak = 1  # 取消破坏
    PlayerActionStopBreak = 2  # 停止破坏
    PlayerActionGetUpdatedBlock = 3  # 获取更新的方块
    PlayerActionDropItem = 4  # 丢弃物品
    PlayerActionStartSleeping = 5  # 开始睡觉
    PlayerActionStopSleeping = 6  # 停止睡觉
    PlayerActionRespawn = 7  # 重生
    PlayerActionJump = 8  # 跳跃
    PlayerActionStartSprint = 9  # 开始冲刺
    PlayerActionStopSprint = 10  # 停止冲刺
    PlayerActionStartSneak = 11  # 开始潜行
    PlayerActionStopSneak = 12  # 停止潜行
    PlayerActionCreativePlayerDestroyBlock = 13  # 创造模式下玩家摧毁方块
    PlayerActionDimensionChangeDone = 14  # 维度切换完成
    PlayerActionStartGlide = 15  # 开始滑翔
    PlayerActionStopGlide = 16  # 停止滑翔
    PlayerActionBuildDenied = 17  # 建造被拒绝
    PlayerActionCrackBreak = 18  # 破坏进度
    PlayerActionChangeSkin = 19  # 更换皮肤
    PlayerActionSetEnchantmentSeed = 20  # 设置附魔种子
    PlayerActionStartSwimming = 21  # 开始游泳
    PlayerActionStopSwimming = 22  # 停止游泳
    PlayerActionStartSpinAttack = 23  # 开始旋转攻击
    PlayerActionStopSpinAttack = 24  # 停止旋转攻击
    PlayerActionStartBuildingBlock = 25  # 开始建造方块
    PlayerActionPredictDestroyBlock = 26  # 预测破坏方块
    PlayerActionContinueDestroyBlock = 27  # 继续破坏方块
    PlayerActionStartItemUseOn = 28  # 开始使用物品
    PlayerActionStopItemUseOn = 29  # 停止使用物品

class SubPacket_CmdOutputMsg:
    """命令输出消息子包构建"""
    Success: bool
    Message: str
    Parameters: list[str]

    def __init__(self, pkt: dict):
        self.Success = pkt["Success"]
        self.Parameters = pkt["Parameters"]
        self.Message = pkt["Message"]


class SubPacket_CmdOrigin:
    "命令来源子包构建"
    Origin: int
    UUID: str
    RequestID: str
    PlayerUniqueID: int

    def __init__(self, pkt: dict):
        self.Origin = pkt["Origin"]
        self.UUID = pkt["UUID"]
        self.RequestID = pkt["RequestID"]
        self.PlayerUniqueID = pkt["PlayerUniqueID"]


class Packet_CommandOutput:
    "命令输出包构建"
    CommandOrigin: SubPacket_CmdOrigin
    OutputType: int
    SuccessCount: int
    OutputMessages: list[SubPacket_CmdOutputMsg]
    as_dict: dict

    def __init__(self, pkt: dict):
        self.as_dict = pkt
        self.CommandOrigin = SubPacket_CmdOrigin(pkt["CommandOrigin"])
        self.OutputMessages = [
            SubPacket_CmdOutputMsg(imsg) for imsg in pkt["OutputMessages"]
        ]
        self.SuccessCount = pkt["SuccessCount"]
        self.OutputType = pkt["OutputType"]



"""
## ToolDelta: NEMC Rental Server R-Plugin Loader ##
ToolDelta: 网易我的世界手机版 租赁服机器人式插件加载器
- 进入 ToolDelta 界面:\n
   ```python
   from tooldelta import client_title\n
   client_title()
   ```
- 快速启动 ToolDelta:\n
   ```python
   from tooldelta import start_tool_delta\n
   start_tool_delta()
   ```
- 快速创建插件所需库环境\n
   ```python
   from tooldelta import *
   ```
"""
from .color_print import Print
from .plugin_load.PluginGroup import Plugin
from .starter import plugin_group as plugins
from .starter import start_tool_delta, safe_jump
from .starter import tooldelta
from .frame import Config, Utils, ToolDelta, GameCtrl
from .launch_options import client_title
from .plugin_load import TYPE_CHECKING

# 重定向
Builtins = Utils
Frame = ToolDelta


"""自定义常用 URL 方法"""
import os
import re
import shutil
import socket
import time
from typing import Union
from concurrent.futures import ThreadPoolExecutor, as_completed
import pyspeedtest
import requests

from .get_tool_delta_version import get_tool_delta_version
from .color_print import Print

# 使用方法 mirror_github[value: int].format(url: str)
mirror_github = ["https://tdload.tblstudio.cn/{}", "https://hub.gitmirror.com/{}",
                 "https://gh.con.sh/{}", "https://mirror.ghproxy.com/{}"]

def format_mirror_url(url: str) -> list:
    """填充 url 到镜像 url 列表

    Args:
        url (str): 原始 URL

    Returns:
        list: 填充原始 url 后的镜像列表
    """
    mir_url: list = []
    for mirror in mirror_github:mir_url.append(mirror.format(url))
    return mir_url

def githubdownloadurl_to_rawurl(url: str) -> str:
    """将 GitHub 下载链接转换为原始链接

    Args:
        url (str): 原始链接

    Returns:
        str: 原始链接
    """
    try:
        if url.startswith("https://github.com/"):
            return requests.head(url, allow_redirects=True).url
        else:
            return url
    except:
        return url

def progress_bar(
    current: float | int, total: float | int, length: int | float = 20, color1: str = "§f", color2: str = "§b"
) -> str:
    """执行进度条

    Args:
        current (float | int): 当前进度值
        total (float | int): 总进度值
        length (int): 进度条长度.
        color1 (str): 进度条颜色 1.
        color2 (str): 进度条颜色 2.

    Returns:
        str: 格式化后的进度条字符串
    """
    pc = round(current / total * length)
    return Print.colormode_replace(
        color1 + " " * pc + color2 + " " * (20 - pc) + "§r ", 7
    )


def download_progress_bar(
    current_bytes: int, total_bytes: int, speed: float = 0
) -> None:
    """构建下载进度条

    Args:
        current_bytes (int): 当前已下载的字节数
        total_bytes (int): 文件总字节数
        speed ( float): 下载速度.
    """
    progressBar = progress_bar(current_bytes, total_bytes)
    b = f"{progressBar} {pretty_kb(current_bytes)}B / {pretty_kb(total_bytes)}B"
    if speed != 0:
        b += f" ({pretty_kb(speed)}B/s)    "
    with Print.lock:Print.print_with_info(b, "§a 下载 ", need_log=False, end="\r")


def pretty_kb(n: float) -> str:
    """将字节数转换为可读性更好的字符串表示形式

    Args:
        n (float): 字节数

    Returns:
        str: 可读性更好的字符串表示形式
    """
    if n >= 1048576:
        return f"{round(n / 1048576, 2)}M"
    if n >= 1024:
        return f"{round(n / 1024, 2)}K"
    return f"{round(n, 1)}"


def is_common_text_file(url_path: str) -> bool:
    """判断是否为常见的文本文件.

    Args:
        url_path (str): 文件路径

    Returns:
        bool: 是否为常见的文本文件
    """
    return any(
        url_path.endswith(i)
        for i in [".txt", ".yml", ".md", ".xml", ".html", ".py", ".h", ".c", ".pyi", ".js", ".json"]
    )


def get_file_size(url: str) -> Union[int, None]:
    """获取文件大小

    Args:
        url (str): 网址

    Returns:
        Union[int, None]: 文件大小（单位：字节）
    """
    response = requests.head(url, timeout=10)
    if "Content-Length" in response.headers:
        file_size = int(response.headers["Content-Length"])
        return file_size


def download_file_chunk(url: str, start_byte: int, end_byte: int, save_dir: str) -> int:
    """下载文件的代码块

    Args:
        url (str): 文件的 URL 地址
        start_byte (int): 下载的起始字节位置
        end_byte (int): 下载的结束字节位置
        save_dir (str): 文件保存的目录

    Returns:
        int: 已下载的字节数
    """
    headers = {"Range": f"bytes={start_byte}-{end_byte}"}
    response = requests.get(url, headers=headers, stream=True, timeout=10)
    response.raise_for_status()
    with open(save_dir + ".tmp", "rb+") as dwnf:
        dwnf.seek(start_byte)
        downloaded_bytes = 0
        for chunk in response.iter_content(chunk_size=8192):
            dwnf.write(chunk)
            downloaded_bytes += len(chunk)
        return downloaded_bytes


def download_file_singlethreaded(
    url: str, save_dir: str
) -> None:
    """下载单个文件

    Args:
        url (str): 文件的 URL 地址
        save_dir (str): 文件保存的目录
        ignore_warnings (bool, optional): 是否忽略警告
    """
    with requests.get(url, stream=True, timeout=10) as res:
        res.raise_for_status()
        filesize = get_file_size(url)
        chunk_size = 8192
        useSpeed: float = 0
        # nowsize: 当前已下载的字节数
        nowsize: int = 0
        lastime: float = 1
        with open(save_dir + ".tmp", "wb") as dwnf:
            for chk in res.iter_content(chunk_size):
                nowtime = time.time()
                if nowtime != lastime:
                    useSpeed = chunk_size / (nowtime - lastime)
                    lastime = nowtime
                nowsize += len(chk)
                dwnf.write(chk)
                if nowsize % 81920 == 0 and filesize:  # 每下载 10 个数据块更新一次进度
                    download_progress_bar(nowsize, filesize, useSpeed)
        shutil.move(save_dir + ".tmp", save_dir)


def download_unknown_file(url: str, save_dir: str) -> None:
    """下载未知文件

    Args:
        url (str): 文件的 URL 地址
        save_dir (str): 文件保存的目录
    """
    # 鉴于 Content-Length 不一定表示文件原始大小，二进制文件与文本文件需要分开下载
    # 否则显示的下载条会异常
    resp = requests.get(url, timeout=10)
    resp.raise_for_status()
    if is_common_text_file(save_dir):
        # 文本文件，体积可能不大
        with open(save_dir, "wb") as f:
            f.write(resp.content)
    download_file_singlethreaded(url, save_dir)


def test_site_latency(Da: dict) -> list:
    """测试网站延迟

    Args:
        Da (dict): 包含 URL 和镜像 URL 的字典

    Returns:
        list: 按延迟排序的 URL 和延迟时间的元组列表
    """
    tmp_speed = {}
    urls = [Da["url"]] + Da["mirror_url"]

    with ThreadPoolExecutor() as executor:
        futures = [executor.submit(measure_latencyt, url) for url in urls]

        for future, url in zip(as_completed(futures), urls):
            try:
                latency = future.result(timeout=5)
                if latency != -1:
                    tmp_speed[url] = latency
            except Exception as e:
                Print.print_war(f"Error measuring latency for {url}: {e}")

    return sorted(tmp_speed.items(), key=lambda x: x[1], reverse=True)


def measure_latencyt(url: str) -> float:
    """测量延迟

    Args:
        url (str): 网址

    Raises:
        ValueError: 无效的网址

    Returns:
        float: 延迟时间
    """
    try:
        # 提取域名
        domain = re.search(
            r"(?<=http[s]://)[.\w-]*(:\d{1,8})?((?=/)|(?!/))", url)
        if isinstance(domain, type(None)):  # 如果没有匹配到域名
            raise ValueError("Invalid URL")
        st = pyspeedtest.SpeedTest(domain.group())  # 传入域名
        download_speed = st.download()
        return download_speed
    except Exception as e:
        Print.print_war(f"Error measuring latency for {url}: {e}")
    return -1.0  # 返回 -1 表示测速失败


def get_free_port(start: int = 2000, end: int = 65535) -> int:
    """获取空闲端口号

    Args:
        start (int, optional): 起始端口号.
        end (int, optional): 结束端口号.

    Raises:
        Exception: 未找到空闲端口

    Returns:
        int: 空闲端口号
    """
    for port in range(start, end):
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            try:
                s.bind(("localhost", port))
                return port
            except OSError:
                Print.print_war(f"端口 {port} 正被占用，跳过")
    raise ValueError(f"未找到空闲端口 ({start}~{end})")


def check_update() -> None:
    """检查更新"""
    try:
        latest_version: str = requests.get(
            "https://tdload.tblstudio.cn/https://api.github.com/repos/ToolDelta/ToolDelta/releases/latest", timeout=5).json()["tag_name"]
        current_version = ".".join(
            map(str, get_tool_delta_version()[:3]))

        if not latest_version.replace(".", "") <= current_version.replace(".", ""):
            # Print.print_suc(f"当前为最新版本 -> v{current_version}，无需更新")
            Print.print_load(
                f"检测到最新版本 {current_version} -> {latest_version}，请及时更新！"
            )
    except KeyError:
        Print.print_war("获取最新版本失败，请检查网络连接")

def if_token() -> None:
    """检查路径下是否有 fbtoken，没有就提示输入

    Raises:
        SystemExit: 未输入 fbtoken
    """
    if not os.path.isfile("fbtoken"):
        Print.print_inf(
            "请到对应的验证服务器官网下载 FBToken，并放在本目录中，或者在下面输入 fbtoken"
        )
        fbtoken = input(Print.fmt_info("请输入 fbtoken: ", "§b 输入 "))
        if fbtoken:
            with open("fbtoken", "w", encoding="utf-8") as f:
                f.write(fbtoken)
        else:
            Print.print_err("未输入 fbtoken, 无法继续")
            raise SystemExit


def fbtokenFix():
    """修复 fbtoken 里的换行符"""
    with open("fbtoken", "r", encoding="utf-8") as f:
        token = f.read()
        if "\n" in token:
            Print.print_war("fbtoken 里有换行符，会造成 fb 登陆失败，已自动修复")
            with open("fbtoken", "w", encoding="utf-8") as f:
                f.write(token.replace("\n", ""))



"启动页面"
import os
import signal
import traceback
from .color_print import Print
from .starter import start_tool_delta
from .frame import ToolDelta
from .plugin_manager import plugin_manager
from .plugin_market import market
from .sys_args import sys_args_to_dict, print_help


def signal_handler(*_) -> None:
    """排除信号中断"""
    return


signal.signal(signal.SIGINT, signal_handler)


def client_title() -> None:
    "选择启动模式"
    try:
        if "h" in sys_args_to_dict() or "help" in sys_args_to_dict():
            print_help()
            raise SystemExit
        launch_mode = sys_args_to_dict()
        if launch_mode.get("l"):
            if not isinstance(launch_mode["l"], str):
                raise ValueError("启动模式参数不合法")
            r = launch_mode["l"]
        else:
            Print.clean_print("§a请选择启动模式§6(使用启动参数 -l <启动模式> 可以跳过该页面):")
            Print.clean_print("1 - §b启动 ToolDelta")
            Print.clean_print("2 - §d打开 ToolDelta 插件管理器")
            Print.clean_print("3 - §d打开 ToolDelta 插件市场")
            Print.clean_print("4 - §a修改 ToolDelta 启动配置")
            r = input("请选择：").strip()
        match r:
            case "1":
                start_tool_delta()
            case "2":
                plugin_manager.manage_plugins()
            case "3":
                market.enter_plugin_market()
            case "4":
                ToolDelta.change_config()
            case _:
                Print.clean_print("§c不合法的启动模式: " + r)
    except (EOFError, SystemExit):
        pass
    except Exception:
        Print.print_err("ToolDelta 运行过程中出现问题：" + traceback.format_exc())



import websockets
import asyncio
import uuid
import ujson
from websockets.legacy.server import WebSocketServerProtocol

class Server:
    def __init__(self, port: int):
        self.port = port
        self.is_conn = False

    def set_event_cbs(
        self,
        on_inject,
        on_data_recv
    ):
        self.on_inject = on_inject
        self.on_data_recv = on_data_recv

    async def wait_connection(self):
        async with websockets.serve(self.connect_once, "0.0.0.0", self.port) as ws_server:
            self.ws_server = ws_server
            await asyncio.Future()

    async def connect_once(self, ws: WebSocketServerProtocol):
        if self.is_conn:
            await ws.close()
        else:
            self.ws = ws
            await self.handle(ws)

    async def handle(self, ws: WebSocketServerProtocol):
        while 1:
            self.on_data_recv(await ws.recv())

    async def send(self, request: dict):
        await self.ws.send(ujson.dumps(request, ensure_ascii=False))

    async def subscribe(self, event_name: str):
        await self.send({
            'body': {'eventName': event_name},
            'header': build_header('subscribe')
        })

    async def unsubscribe(self, event_name):
        await self.send({
            'body': {'eventName': event_name},
            'header': build_header('unsubscribe')
        })

    async def sendcmd(self, command: str):
        request = {
            'body': {'commandLine': command},
            'header': build_header('commandRequest')
        }
        await self.send(request)
        return request["header"]["requestId"]

def build_header(message_purpose, request_id=None):
    if not request_id:
        request_id = str(uuid.uuid4())
    return {
        'requestId': request_id,
        'messagePurpose': message_purpose,
        'version': '1',
        'messageType': 'commandRequest',
    }


"""
游戏交互实用方法

Methods:
    getTarget (sth, timeout): 获取符合目标选择器实体的列表
    getPos (targetNameToGet, timeout): 获取目标玩家的详细位置信息
    getItem (targetName, itemName, itemSpecialID): 获取玩家背包内指定的物品的数量
    getPosXYZ (player, timeout=30): 获取玩家的简略坐标值，并以坐标三元元组返回
    getScore (scoreboardNameToGet, targetNameToGet): 获取计分板分数
    isCmdSuccess (cmd: str, timeout=30): 获取命令执行成功与否的状态
"""
from typing import TYPE_CHECKING
from .color_print import Print
from .packets import Packet_CommandOutput

if TYPE_CHECKING:
    from tooldelta import ToolDelta, GameCtrl

game_ctrl: "GameCtrl" = None # type: ignore

# set_frame

def _check_gamectrl_avali():
    "检查 GameCtrl 是否可用"
    if game_ctrl is None:
        raise ValueError("GameControl 不可用")

def _set_frame(frame: "ToolDelta"):
    "载入系统框架"
    global game_ctrl
    game_ctrl = frame.get_game_control()

# utils

def getTarget(sth: str, timeout: int = 5) -> list:
    """
    获取符合目标选择器实体的列表

    参数:
        sth: 目标选择器
        timeout: 超时时间，默认为 5 秒
    异常:
        ValueError: 指令返回超时，或者无法获取目标
    """
    _check_gamectrl_avali()
    if not sth.startswith("@"):
        raise ValueError("我的世界目标选择器格式错误 (getTarget 必须使用目标选择器)")
    result = game_ctrl.sendcmd_with_resp(f"/testfor {sth}", timeout)
    if result.SuccessCount:
        result = result.OutputMessages[0].Parameters[0]
        return result.split(", ")
    else:
        if result.OutputMessages[0].Message == "commands.generic.syntax":
            raise ValueError(f"getTarget 目标选择器表达式错误: {sth}")
        else:
            return []

def getPos(target: str, timeout: float | int = 5) -> dict:
    """获取目标玩家的详细位置信息

    参数:
        targetNameToGet: 目标玩家的名称
        timeout: 超时时间（秒）。默认为 5 秒

    异常:
        ValueError: 当目标玩家不存在时抛出该异常
        ValueError: 当获取位置信息失败时抛出该异常
        AttributeError: 当获取玩家 UUID 失败时抛出该异常
    """
    _check_gamectrl_avali()
    if target not in game_ctrl.allplayers and not target.startswith("@") and target != game_ctrl.bot_name:
        raise ValueError(f'玩家 "{target}" 不存在')
    result = game_ctrl.sendcmd_with_resp(f'/querytarget @a[name="{target}"]', timeout)
    if not result.OutputMessages[0].Success:
        raise ValueError(f"无法获取坐标信息：{result.OutputMessages[0].Message}")
    parameter = result.OutputMessages[0].Parameters[0]
    if isinstance(parameter, str):
        raise ValueError("无法获取坐标信息：" + parameter)
    result = {}
    if game_ctrl.players_uuid is None:
        raise AttributeError("无法获取玩家 UUID 表")
    targetName = target
    x = (
        parameter[0]["position"]["x"]
        if parameter[0]["position"]["x"] >= 0
        else parameter[0]["position"]["x"] - 1
    )
    y = parameter[0]["position"]["y"] - 1.6200103759765
    z = (
        parameter[0]["position"]["z"]
        if parameter[0]["position"]["z"] >= 0
        else parameter[0]["position"]["z"] - 1
    )
    position = {
        "x": float(f"{x:.2f}"),
        "y": float(f"{y:.2f}"),
        "z": float(f"{z:.2f}"),
    }
    dimension = parameter[0]["dimension"]
    yRot = parameter[0]["yRot"]
    result[targetName] = {
        "dimension": dimension,
        "position": position,
        "yRot": yRot,
    }
    if target == "@a":
        return result
    if len(result) != 1:
        raise ValueError("获取坐标失败")
    if target.startswith("@a"):
        return list(result.values())[0]
    return result[target]

def getItem(target: str, itemName: str, itemSpecialID: int = -1) -> int:
    """
    获取玩家背包内指定的物品的数量
    参数:
        targetName (str): 玩家选择器 / 玩家名
        itemName (str): 物品 ID
        itemSpecialID (int): 物品特殊值, 默认值 -1
    """
    if (
        (target not in game_ctrl.allplayers)
        and (target != game_ctrl.bot_name)
        and (not target.startswith("@a"))
    ):
        raise ValueError("未找到目标玩家")
    result: Packet_CommandOutput = game_ctrl.sendcmd_with_resp(
        f"/clear {target} {itemName} {itemSpecialID} 0")
    if result.OutputMessages[0].Message == "commands.generic.syntax":
        raise ValueError("物品 ID 错误")
    if result.OutputMessages[0].Message == "commands.clear.failure.no.items":
        return 0
    return int(result.OutputMessages[0].Parameters[1])

def getPosXYZ(player, timeout: int | float = 30) -> tuple[float, float, float]:
    """
    获取玩家的简略坐标值，并以坐标三元元组返回
    参数:
        player (str): 玩家名
        timeout (int): 最长超时时间
    返回:
        tuple[float, float, float]
    """
    res = getPos(player, timeout=timeout)["position"]
    return res["x"], res["y"], res["z"]

def getMultiScore(scoreboardNameToGet: str, targetNameToGet: str) -> int | dict:
    """
    获取单个或多个计分板分数项
    参数:
        scoreboardNameToGet: 计分板名
        targetNameToGet: 获取分数的对象/目标选择器
    返回:
        分数：int
    异常:
        ValueError: 无法获取分数
    """
    resultList = game_ctrl.sendcmd_with_resp(
        f"/scoreboard players list {targetNameToGet}"
    ).OutputMessages
    result = {}
    result2 = {}
    for i in resultList:
        Message = i.Message
        if Message == r"commands.scoreboard.players.list.player.empty":
            continue
        if Message == r"§a%commands.scoreboard.players.list.player.count":
            targetName = i.Parameters[1][1:]
        elif Message == "commands.scoreboard.players.list.player.entry":
            if targetName == "commands.scoreboard.players.offlinePlayerName":
                continue
            scoreboardName = i.Parameters[2]
            targetScore = int(i.Parameters[0])
            if targetName not in result:
                result[targetName] = {}
            result[targetName][scoreboardName] = targetScore
            if scoreboardName not in result2:
                result2[scoreboardName] = {}
            result2[scoreboardName][targetName] = targetScore
    if not (result or result2):
        raise Exception("获取计分板分数失败")
    try:
        if targetNameToGet == "*" or targetNameToGet.startswith("@"):
            return result2[scoreboardNameToGet]
        if scoreboardNameToGet == "*":
            return result[targetNameToGet]
        return result[targetNameToGet][scoreboardNameToGet]
    except KeyError as err:
        raise Exception(f"获取计分板分数失败：{err}")

def getScore(scb_name: str, target: str, timeout = 30) -> int:
    _check_gamectrl_avali()
    if target == "*" or scb_name == "*":
        raise ValueError("在此处无法使用 通配符 作为计分板分数获取目标")
    resp = game_ctrl.sendcmd_with_resp(f"/scoreboard players test {target} {scb_name} 0 0", timeout).OutputMessages[0]
    if resp.Message == "commands.scoreboard.objectiveNotFound":
        raise ValueError(f"计分板 {scb_name} 未找到")
    elif resp.Message == "commands.scoreboard.players.list.player.empty":
        raise ValueError(f"计分板项或玩家 {target} 未找到")
    elif resp.Message == "commands.scoreboard.players.score.notFound":
        raise ValueError(f"计分板项或玩家 {target} 在此计分板没有分数")
    elif not resp.Success:
        raise ValueError(f"计分板分数获取失败: {resp.Message}")
    return int(resp.Parameters[0])

def isCmdSuccess(cmd: str, timeout=30):
    """
    获取命令执行成功与否的状态
    参数:
        cmd: MC 指令
        timeout: 超时时间
    返回:
        命令执行是否成功：bool
    """
    res = game_ctrl.sendcmd_with_resp(cmd, timeout).SuccessCount
    return bool(res)


import requests
import platform
import os
import json

from tooldelta.color_print import Print
from tooldelta.urlmethod import download_file_singlethreaded
from tooldelta.sys_args import sys_args_to_dict
from tooldelta.cfg import Config
from tooldelta import constants

def download_libs() -> bool:
    """根据系统架构和平台下载所需的库。"""
    if "no-download-libs" in sys_args_to_dict().keys():
        Print.print_war("将不会进行依赖库的下载和检测更新。")
        return True
    cfgs = Config.get_cfg("ToolDelta基本配置.json", constants.LAUNCH_CFG_STD)
    is_mir: bool = cfgs["是否使用github镜像"]
    if is_mir:
        mirror_src = constants.TDSPECIFIC_MIRROR + \
            "/https://raw.githubusercontent.com/ToolDelta/ToolDelta/main/"
        depen_url = constants.TDSPECIFIC_MIRROR + \
            "/https://raw.githubusercontent.com/ToolDelta/DependencyLibrary/main/"
    else:
        mirror_src = "https://raw.githubusercontent.com/ToolDelta/ToolDelta/main/"
        depen_url = "https://raw.githubusercontent.com/ToolDelta/DependencyLibrary/main/"
    try:
        require_depen = json.loads(
            requests.get(
            f"{mirror_src}require_files.json", timeout=5
            ).text
        )
    except Exception as err:
        Print.print_err(f"获取依赖库表出现问题：{err}")
        return False
    sys_machine = platform.machine().lower()
    if sys_machine == "x86_64":
        sys_machine = "amd64"
    elif sys_machine == "aarch64":
        sys_machine = "arm64"
    if "TERMUX_VERSION" in os.environ:
        sys_info_fmt: str = f"Android:{sys_machine.lower()}"
    else:
        sys_info_fmt: str = f"{platform.uname().system}:{sys_machine.lower()}"
    source_dict: list[str] = require_depen[sys_info_fmt]
    commit_remote = requests.get(
        f"{depen_url}commit", timeout=5
    ).text
    commit_file_path = os.path.join(
        os.getcwd(), "tooldelta", "neo_libs", "commit")
    replace_file = False
    if os.path.isfile(commit_file_path):
        with open(commit_file_path, "r", encoding="utf-8") as f:
            commit_local = f.read()
        if commit_local != commit_remote:
            Print.print_war("依赖库版本过期，将重新下载")
            replace_file = True
    else:
        replace_file = True
    for v in source_dict:
        pathdir = os.path.join(os.getcwd(), "tooldelta", "neo_libs", v)
        url = depen_url + v
        if not os.path.isfile(pathdir) or replace_file:
            Print.print_with_info(f"正在下载依赖库 {pathdir} ...", "§a 下载 §r")
            try:
                download_file_singlethreaded(url, pathdir)
            except Exception as err:
                Print.print_err(f"下载依赖库出现问题：{err}")
                return False
    if replace_file:
        # 写入 commit_remote，文字写入
        with open(commit_file_path, "w", encoding="utf-8") as f:
            f.write(commit_remote)
        Print.print_suc("已完成依赖更新！")
    return True


"认证模块"

import getpass
import hashlib
from typing import Any
import requests
import ujson as json
from tooldelta import constants
from .color_print import Print


def liliya_login() -> str:
    """登录咕咕账号

    Raises:
        requests.exceptions.RequestException: 登录失败
    """
    hash_obj = hashlib.sha256()
    username = input(Print.fmt_info("请输入账号：", "§6 账号 "))
    hash_obj.update(
        getpass.getpass(Print.fmt_info("请输入密码 (已隐藏):", "§6 密码 ")).encode()
    )
    password = hash_obj.hexdigest()
    auth_key = requests.get(url=constants.GUGU_APIS[1], timeout=5).text
    repo = requests.post(
        url=constants.GUGU_APIS[0],
        data=json.dumps(
            {
                "client_public_key": "",
                "server_code": "::DRY::",
                "server_passcode": "::DRY::",
                "username": username,
                "password": password,
            },
            ensure_ascii=False,
        ),
        headers={
            "Content-Type": "application/json",
            "Authorization": f"Bearer {auth_key}",
        }, timeout=5
    )
    repo_text: dict[str, Any] = json.loads(repo.text)
    repo_message: str = repo_text["message"]
    repo_success: bool = repo_text["success"]
    if repo.status_code != 200:
        Print.print_war(
            f"请求 Api 接口失败，将自动使用 Token 登陆！状态码:{repo.status_code}，返回值:{repo.text}"
        )
        raise requests.exceptions.RequestException("请求 Api 接口失败", repo)
    if not repo_success:
        if "无效的用户中心用户名或密码" in repo_message:
            Print.print_war("登录失败，无效的用户名或密码！")
            raise requests.exceptions.RequestException("无效的用户名或密码")
        Print.print_war(f"登录失败，原因：{repo_message}")
        raise requests.exceptions.RequestException("登录无效", repo)
    token = repo_text["token"]
    return token


def fbuc_login() -> str:
    """登录 FastBuilder 账号

    Raises:
        requests.exceptions.RequestException: 登录失败
    """
    hash_obj = hashlib.sha256()
    username = input(Print.fmt_info("请输入账号：", "§6 账号 "))
    hash_obj.update(
        getpass.getpass(
            Print.fmt_info("请输入密码 (已隐藏):", "§6 密码 ")
        ).encode()
    )
    password = hash_obj.hexdigest()
    mfa_code = getpass.getpass(
        Print.fmt_info(
            "请输入双重验证码 (已隐藏)(如未设置请直接回车):", "§6 MFA  "
        )
    )
    auth_key = requests.get(url=constants.FB_APIS[1], timeout=5).text
    repo = requests.post(
        url=constants.FB_APIS[3],
        data=json.dumps(
            {
                "username": username,
                "password": password,
                "mfa_code": mfa_code,
            },
            ensure_ascii=False,
        ),
        headers={
            "Content-Type": "application/json",
            "authorization": f"Bearer {auth_key}",
        },
        timeout=5
    )
    repo_text: dict[str, Any] = json.loads(repo.text)
    repo_message: str = repo_text["message"]
    repo_success: bool = repo_text["success"]
    if repo.status_code != 200:
        Print.print_war(
            f"请求 Api 接口失败，将自动使用 Token 登陆！状态码:{repo.status_code}，返回值:{repo.text}"
        )
        raise requests.exceptions.RequestException("请求 Api 接口失败", repo)

    if not repo_success:
        if "Invalid username, password, or MFA code." in repo_message:
            raise requests.exceptions.RequestException("无效的用户名、密码或 MFA 代码")
        raise requests.exceptions.RequestException("登录无效", repo)
    # 获取 token 前缀
    repo = requests.get(
        url=constants.FB_APIS[2],
        data=json.dumps({"with_prefix": constants.FB_APIS[4]}),
        timeout=5,
        headers={
            "Content-Type": "application/json",
            "authorization": f"Bearer {auth_key}",
        },
    )
    if repo.status_code != 200:
        Print.print_war(
            f"无法获取 tokenx 信息！状态码:{repo.status_code}，返回值:{repo.text}"
        )
        raise requests.exceptions.RequestException("请求 Api 接口失败", repo)
    with_perfix: dict[str, Any] = json.loads(repo.text)
    token_request = requests.get(
        url=constants.FB_APIS[4] + with_perfix["get_phoenix_token"],
        data=json.dumps({"secret": f"{auth_key}"}),
        timeout=5,
        headers={
            "Content-Type": "application/json",
            "authorization": f"Bearer {auth_key}",
        },
    )
    if token_request.status_code != 200:
        Print.print_war(
            f"无法获取 tokenx 信息！状态码:{token_request.status_code}，返回值:{token_request.text}"
        )
        raise requests.exceptions.RequestException("请求 Api 接口失败", token_request)
    return token_request.text



"解析启动参数"
import sys


def sys_args_to_dict(argv: list[str] = sys.argv)-> dict[str, str | None]:
    """将启动参数转换为字典

    Args:
        argv (list[str], optional): 启动参数

    Returns:
        dict[str, str | None]: 启动参数字典
    """
    # 为什么不用 argsparser? 因为这样可以支持混合型启动参数
    d: dict[str, str | None] = {}
    i = 0
    while i < len(argv):
        arg = argv[i]
        if arg.startswith("-"):
            arg = arg.strip("--").strip("-")
            try:
                val = argv[i + 1]
                if val.startswith("-"):
                    val = None
                else:
                    i += 1
            except IndexError:
                val = None
            d[arg] = val
        i += 1
    return d

def print_help():
    print("启动参数帮助：")
    print("    -help  查看帮助 (当前页)")
    print("    -no-update-check  禁用自动更新")
    print("    -no-download-libs  禁用自动更新依赖库")
    print("    -l <启动模式>  按照特定的启动模式，跳过启动页面")


"""客户端启动器框架"""

import os
import threading
import shlex
import time
import random
import subprocess
import platform
from typing import Callable, Optional
import ujson as json
import requests
import tooldelta

from tooldelta import constants
from .neo_libs import file_download as neo_fd, neo_conn
from .bews_libs import ws_con
from .cfg import Cfg
from .utils import Utils
from .color_print import Print
from .sys_args import sys_args_to_dict
from .packets import Packet_CommandOutput
from .urlmethod import download_file_singlethreaded, get_free_port

Config = Cfg()


class SysStatus:
    """系统状态码

    LOADING: 启动器正在加载
    LAUNCHING: 启动器正在启动
    RUNNING: 启动器正在运行
    NORMAL_EXIT: 正常退出
    FB_LAUNCH_EXC: FastBuilder 启动异常
    CRASHED_EXIT: 启动器崩溃退出
    NEED_RESTART: 需要重启
    """
    LOADING = 100
    LAUNCHING = 101
    RUNNING = 102
    NORMAL_EXIT = 103
    FB_LAUNCH_EXC = 104
    CRASHED_EXIT = 105
    NEED_RESTART = 106
    launch_type = "None"


class StandardFrame:
    """提供了标准的启动器框架，作为 ToolDelta 和游戏交互的接口"""
    launch_type = "Original"

    def __init__(self) -> None:
        """实例化启动器框架

        Args:
            serverNumber (int): 服务器号
            password (str): 服务器密码
            fbToken (str): 验证服务器 Token
            auth_server_url (str): 验证服务器地址
        """
        self.system_type = platform.uname().system
        self.inject_events: list = []
        self.packet_handler: Optional[Callable] = lambda pckType, pck: None
        self.need_listen_packets: set[int] = {9, 63, 79}
        self._launcher_listener: Callable
        self.exit_event = threading.Event()
        self.status: int = SysStatus.LOADING

    def init(self):
        "初始化启动器框架"

    def add_listen_packets(self, *pcks: int) -> None:
        """添加需要监听的数据包"""
        for i in pcks:
            self.need_listen_packets.add(i)

    def launch(self) -> None:
        """启动器启动

        Raises:
            SystemError: 无法启动此启动器
        """
        raise NotImplementedError

    def listen_launched(self, cb: Callable) -> None:
        """设置监听启动器启动事件"""
        self._launcher_listener = cb

    def get_players_and_uuids(self) -> None:
        """获取玩家名和 UUID"""
        raise NotImplementedError

    def get_bot_name(self) -> str:
        """获取机器人名字"""
        raise NotImplementedError

    def update_status(self, new_status: int) -> None:
        """更新启动器状态

        Args:
            new_status (int): 新的状态码
        """
        self.status = new_status
        if new_status == SysStatus.NORMAL_EXIT:
            tooldelta.safe_jump(out_task=True)
            self.exit_event.set()  # 设置事件，触发等待结束
        if new_status == SysStatus.CRASHED_EXIT:
            tooldelta.safe_jump(out_task=False)
            self.exit_event.set()

    def sendcmd(self, cmd: str, waitForResp: bool = False, timeout: int | float = 30) -> Optional[Packet_CommandOutput]:
        """以玩家身份发送命令

        Args:
            cmd (str): 命令
            waitForResp (bool, optional): 是否等待结果
            timeout (int | float, optional): 超时时间

        Raises:
            NotImplementedError: 未实现此方法

        Returns:
            Optional[Packet_CommandOutput]: 返回命令结果
        """
        raise NotImplementedError

    def sendwscmd(self, cmd: str, waitForResp: bool = False, timeout: int | float = 30) -> Optional[Packet_CommandOutput]:
        """以 ws 身份发送命令

        Args:
            cmd (str): 命令
            waitForResp (bool, optional): 是否等待结果
            timeout (int | float, optional): 超时时间

        Raises:
            NotImplementedError: 未实现此方法

        Returns:
            Optional[Packet_CommandOutput]: 返回命令结果
        """
        raise NotImplementedError

    def sendwocmd(self, cmd: str) -> None:
        """以 wo 身份发送命令

        Args:
            cmd (str): 命令

        Raises:
            NotImplementedError: 未实现此方法
        """
        raise NotImplementedError

    def sendPacket(self, pckID: int, pck: str) -> None:
        """发送数据包

        Args:
            pckID (int): 数据包 ID
            pck (str): 数据包内容

        Raises:
            NotImplementedError: 未实现此方法
        """
        raise NotImplementedError

    sendPacketJson = sendPacket

    def is_op(self, player: str) -> bool:
        """检查玩家是否为 OP

        Args:
            player (str): 玩家名

        Raises:
            NotImplementedError: 未实现此方法

        Returns:
            bool: 是否为 OP
        """
        raise NotImplementedError

    def place_command_block_with_nbt_data(self, block_name: str, block_states: str,
                                          position: tuple[int, int, int],
                                          nbt_data: neo_conn.CommandBlockNBTData):
        """在 position 放置方块名为 block_name 且方块状态为 block_states 的命令块，
        同时向该方块写入 nbt_data 所指代的 NBT 数据

        Args:
            block_name (str): 命令块的方块名，如 chain_command_block
            block_states (str): 命令块的方块状态，如 朝向南方 的命令方块表示为 ["facing_direction":3]
            position (tuple[int, int, int]): 命令块应当被放置的位置。三元整数元组从左到右依次对应世界坐标的 X, Y, Z 轴坐标
            nbt_data (neo_conn.CommandBlockNBTData): 该命令块的原始 NBT 数据

        Raises:
            NotImplementedError: 未实现此方法
        """
        raise NotImplementedError


class FrameNeOmg(StandardFrame):
    """使用 NeOmega 框架连接到游戏"""
    launch_type = "NeOmega"

    def __init__(self) -> None:
        """初始化 NeOmega 框架

        Args:
            serverNumber (int): 服务器号
            password (str): 服务器密码
            fbToken (str): 验证服务器 Token
            auth_server (str): 验证服务器地址
        """
        super().__init__()
        self.status = SysStatus.LOADING
        self.launch_event = threading.Event()
        self.neomg_proc = None
        self.serverNumber = None
        self.neomega_account_opt = None
        self.bot_name = ""
        self.omega = neo_conn.ThreadOmega(
            connect_type=neo_conn.ConnectType.Remote,
            address="tcp://localhost:24013",
            accountOption=None
        )

    def init(self):
        res = neo_fd.download_libs()
        if not res:
            raise SystemExit("ToolDelta 因下载库异常而退出")
        neo_conn.load_lib()
        self.status = SysStatus.LAUNCHING
        self.secret_exit_key = ""

    def set_launch_data(self, serverNumber: int, password: str, fbToken: str, auth_server_url: str):
        self.serverNumber = serverNumber
        self.serverPassword = password
        self.fbToken = fbToken
        self.auth_server = auth_server_url
        if self.serverNumber is None:
            self.neomega_account_opt = None
        else:
            self.neomega_account_opt = neo_conn.AccountOptions(
                AuthServer=self.auth_server,
                UserToken=self.fbToken,
                ServerCode=str(self.serverNumber),
                ServerPassword=str(self.serverPassword),
            )
        self.omega = neo_conn.ThreadOmega(
            connect_type=neo_conn.ConnectType.Remote,
            address="tcp://localhost:24013",
            accountOption=self.neomega_account_opt
        )

    def set_omega(self, openat_port: int) -> None:
        """设置 Omega 连接

        Args:
            openat_port (int): 端口号

        Raises:
            SystemExit: 系统退出
        """
        retries = 0
        self.omega.address = f"tcp://localhost:{openat_port}"
        while retries <= 10:
            try:
                self.omega.connect()
                retries = 0
                break
            except Exception as err:
                Print.print_war(f"OMEGA 连接失败，重连：第 {retries} 次：{err}")
                time.sleep(5)
                retries += 1
                if retries > 5:
                    Print.print_err("最大重试次数已超过")
                    raise SystemExit from err

    def start_neomega_proc(self) -> int:
        """启动 NeOmega 进程

        Returns:
            int: 端口号
        """
        free_port = get_free_port(24016)
        sys_machine = platform.uname().machine
        sys_machine = "amd64" if sys_machine == "x86_64" else sys_machine
        access_point_file = (
            f"neomega_{platform.uname().system.lower()}_access_point_{sys_machine}"
        )
        if "TERMUX_VERSION" in os.environ:
            access_point_file = f"neomega_android_access_point_{sys_machine}"
        if platform.system() == "Windows":
            access_point_file += ".exe"
        exe_file_path = os.path.join(
            os.getcwd(), "tooldelta", "neo_libs", access_point_file
        )
        if platform.uname().system.lower() == "linux":
            os.system("chmod +x " + shlex.quote(exe_file_path))
        # 只需要+x 即可
        self.neomg_proc = subprocess.Popen(
            [
                exe_file_path,
                "-server",
                str(self.serverNumber),
                "-T",
                self.fbToken,
                "-access-point-addr",
                f"tcp://localhost:{free_port}",
                "-server-password",
                str(self.serverPassword),
                "-auth-server",
                self.auth_server,
            ],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )
        return free_port

    def _msg_show_thread(self) -> None:
        """显示来自 NeOmega 的信息"""
        if self.neomg_proc is None or self.neomg_proc.stdout is None:
            raise ValueError("NEOMG 进程未启动")
        while True:
            msg_orig = self.neomg_proc.stdout.readline().decode("utf-8").strip("\n")
            if msg_orig in ("", "SIGNAL: exit"):
                with Print.lock:
                    Print.print_with_info(
                        "ToolDelta: NEOMG 进程已结束", "§b NOMG ")
                self.update_status(SysStatus.NORMAL_EXIT)
                return
            if "[neOmega 接入点]: 就绪" in msg_orig:
                self.launch_event.set()
            elif f"STATUS CODE: {self.secret_exit_key}" in msg_orig:
                with Print.lock:
                    Print.print_with_info("§a机器人已退出", "§b NOMG ")
                continue
            with Print.lock:
                Print.print_with_info(msg_orig, "§b NOMG ")

    def make_secret_key(self) -> None:
        """生成退出密钥"""
        self.secret_exit_key = hex(random.randint(10000, 99999))

    def launch(self) -> SystemExit | Exception | SystemError:
        """启动 NeOmega 进程

        returns:
            SystemExit: 正常退出
            Exception: 异常退出
            SystemError: 未知的退出状态
        """
        self.status = SysStatus.LAUNCHING
        openat_port = self.start_neomega_proc()
        Utils.createThread(self._msg_show_thread, usage="显示来自 NeOmega 的信息")
        self.launch_event.wait()
        self.make_secret_key()
        self.set_omega(openat_port)
        self.update_status(SysStatus.RUNNING)
        self.wait_omega_disconn_thread()
        Print.print_suc("已开启接入点进程")
        pcks = [
            self.omega.get_packet_id_to_name_mapping(i)
            for i in self.need_listen_packets
        ]
        self.omega.listen_packets(pcks, self.packet_handler_parent)
        self._launcher_listener()
        Print.print_suc("接入点已就绪！")
        self.exit_event.wait()  # 等待事件的触发
        if self.status == SysStatus.NORMAL_EXIT:
            return SystemExit("正常退出。")
        if self.status == SysStatus.CRASHED_EXIT:
            return Exception("NeOmega 已崩溃")
        return SystemError("未知的退出状态")

    def get_players_and_uuids(self):
        players_uuid = {}
        if self.status != SysStatus.RUNNING:
            raise ValueError("未连接到接入点")
        for i in self.omega.get_all_online_players():
            if i is not None:
                players_uuid[i.name] = i.uuid
            else:
                raise ValueError("未能获取玩家名和 UUID")
        return players_uuid

    def get_bot_name(self) -> str:
        """获取机器人名字

        Returns:
            str: 机器人名字
        """
        self.check_avaliable()
        if not self.bot_name:
            self.bot_name = self.omega.get_bot_name()
        return self.bot_name

    def packet_handler_parent(self, pkt_type: str, pkt: dict) -> None:
        """数据包处理器

        Args:
            pkt_type (str): 数据包类型
            pkt (dict): 数据包内容

        Raises:
            ValueError: 未连接到接入点
        """
        if self.omega is None or self.packet_handler is None:
            raise ValueError("未连接到接入点")
        packetType = self.omega.get_packet_name_to_id_mapping(pkt_type)
        self.packet_handler(packetType, pkt)

    def check_avaliable(self):
        if self.status != SysStatus.RUNNING:
            raise ValueError("未连接到游戏")

    def sendcmd(self, cmd: str, waitForResp: bool = False, timeout: float = 30) -> Optional[Packet_CommandOutput]:
        """以玩家身份发送命令

        Args:
            cmd (str): 命令
            waitForResp (bool, optional): 是否等待结果
            timeout (int | float, optional): 超时时间

        Raises:
            NotImplementedError: 未实现此方法

        Returns:
            Optional[Packet_CommandOutput]: 返回命令结果
        """
        self.check_avaliable()
        if waitForResp:
            res = self.omega.send_player_command_need_response(
                cmd, timeout)
            if res is None:
                raise TimeoutError("指令超时")
            return res
        self.omega.send_player_command_omit_response(cmd)
        return

    def sendwscmd(self, cmd: str, waitForResp: bool = False, timeout: float = 30) -> Optional[Packet_CommandOutput]:
        """以玩家身份发送命令

        Args:
            cmd (str): 命令
            waitForResp (bool, optional): 是否等待结果
            timeout (int | float, optional): 超时时间

        Raises:
            NotImplementedError: 未实现此方法

        Returns:
            Optional[Packet_CommandOutput]: 返回命令结果
        """
        self.check_avaliable()
        if waitForResp:
            res = self.omega.send_websocket_command_need_response(
                cmd, timeout)
            if res is None:
                raise TimeoutError("指令超时")
            return res
        self.omega.send_websocket_command_omit_response(cmd)
        return

    def sendwocmd(self, cmd: str) -> None:
        """以 wo 身份发送命令

        Args:
            cmd (str): 命令

        Raises:
            NotImplementedError: 未实现此方法
        """
        self.check_avaliable()
        self.omega.send_settings_command(cmd)

    def sendPacket(self, pckID: int, pck: str) -> None:
        """发送数据包

        Args:
            pckID (int): 数据包 ID
            pck (str): 数据包内容

        Raises:
            NotImplementedError: 未实现此方法
        """
        self.check_avaliable()
        self.omega.send_game_packet_in_json_as_is(pckID, pck)

    def is_op(self, player: str) -> bool:
        """检查玩家是否为 OP

        Args:
            player (str): 玩家名

        Raises:
            NotImplementedError: 未实现此方法

        Returns:
            bool: 是否为 OP
        """
        self.check_avaliable()
        player_obj = self.omega.get_player_by_name(player)
        if player_obj is None or player_obj.op_permissions_level is None:
            raise ValueError("未能获取玩家对象")
        return player_obj.op_permissions_level > 1

    def place_command_block_with_nbt_data(self, block_name: str, block_states: str,
                                          position: tuple[int, int, int],
                                          nbt_data: neo_conn.CommandBlockNBTData):
        """在 position 放置方块名为 block_name 且方块状态为 block_states 的命令块，
        同时向该方块写入 nbt_data 所指代的 NBT 数据

        Args:
            block_name (str): 命令块的方块名，如 chain_command_block
            block_states (str): 命令块的方块状态，如 朝向南方 的命令方块表示为 ["facing_direction":3]
            position (tuple[int, int, int]): 命令块应当被放置的位置。三元整数元组从左到右依次对应世界坐标的 X, Y, Z 轴坐标
            nbt_data (neo_conn.CommandBlockNBTData): 该命令块的原始 NBT 数据

        Raises:
            NotImplementedError: 未实现此方法
        """
        self.check_avaliable()
        self.omega.place_command_block(neo_conn.CommandBlockPlaceOption(
            X=position[0], Y=position[1], Z=position[2],
            BlockName=block_name, BockState=block_states,
            NeedRedStone=(not nbt_data.ConditionalMode),
            Conditional=nbt_data.ConditionalMode,
            Command=nbt_data.Command,
            Name=nbt_data.CustomName,
            TickDelay=nbt_data.TickDelay,
            ShouldTrackOutput=nbt_data.TrackOutput,
            ExecuteOnFirstTick=nbt_data.ExecuteOnFirstTick
        ))

    @Utils.thread_func("检测 Omega 断开连接线程")
    def wait_omega_disconn_thread(self):
        self.omega.wait_disconnect()
        self.update_status(SysStatus.CRASHED_EXIT)

    sendPacketJson = sendPacket


class FrameNeOmgRemote(FrameNeOmg):
    """远程启动器框架 (使用 NeOmega 框架的 Remote 连接)

    Args:
        FrameNeOmg (FrameNeOmg): FrameNeOmg 框架
    """
    launch_type = "NeOmega Remote"

    def launch(self) -> SystemExit | Exception | SystemError:
        """启动远程启动器框架

        Raises:
            AssertionError: 端口号错误

        Returns:
            SystemExit | Exception | SystemError: 退出状态
        """
        try:
            openat_port = int(sys_args_to_dict().get(
                "access-point-port") or "24020")
            if openat_port not in range(65536):
                raise AssertionError
        except (ValueError, AssertionError):
            Print.print_err("启动参数 -access-point-port 错误：不是 1~65535 的整数")
            raise SystemExit("端口参数错误")
        if openat_port == 0:
            Print.print_war(
                "未用启动参数指定链接 neOmega 接入点开放端口，尝试使用默认端口 24015"
            )
            Print.print_inf("可使用启动参数 -access-point-port 端口 以指定接入点端口。")
            openat_port = 24015
            return SystemExit("未指定端口号")
        Print.print_inf(f"将从端口 {openat_port} 连接至接入点 (等待接入中).")
        self.set_omega(openat_port)
        self.update_status(SysStatus.RUNNING)
        self.wait_omega_disconn_thread()
        Print.print_suc("已连接上接入点进程。")
        pcks = [
            self.omega.get_packet_id_to_name_mapping(i)
            for i in self.need_listen_packets
        ]
        self.omega.listen_packets(pcks, self.packet_handler_parent)
        self._launcher_listener()
        Print.print_suc("接入点已就绪")
        self.exit_event.wait()
        self.update_status(SysStatus.NORMAL_EXIT)
        if self.status == SysStatus.NORMAL_EXIT:
            return SystemExit("正常退出。")
        if self.status == SysStatus.CRASHED_EXIT:
            return Exception("接入点已崩溃")
        return SystemError("未知的退出状态")


class FrameBEConnect(StandardFrame):
    "WIP: Minecraft Bedrock '/connect' 指令所连接的服务端"

    def init(self):
        self.cmd_resp = {}

    def prepare_apis(self):
        ...

    def handler(self, data):
        message_purpose = data['header']['messagePurpose']
        if message_purpose == 'commandResponse':
            request_id = data['header']['requestId']
            if request_id in self.cmd_resp.keys():
                self.cmd_resp[request_id] = Packet_CommandOutput({
                    "CommandOrigin": [],
                    "OutputType": 1,
                })



"""配置文件模块"""

import os
from typing import Any, Union
import ujson

NoneType = type(None)

PLUGINCFG_DEFAULT = {"配置版本": "0.0.1", "配置项": None}
PLUGINCFG_STANDARD_TYPE = {"配置版本": str, "配置项": [type(None), dict]}

def cfg_isinstance_single(obj: Any, typ: type) -> bool:
    if not isinstance(typ, type):
        raise TypeError(f"cfg_isinstance arg 1 must be a type, not {typ}")
    return {
        Cfg.PInt: lambda: isinstance(obj, int) and obj > 0,
        Cfg.NNInt: lambda: isinstance(obj, int) and obj >= 0,
        Cfg.PFloat: lambda: isinstance(obj, float) and obj > 0,
        Cfg.NNFloat: lambda: (isinstance(obj, float) or obj == 0) and obj >= 0,
        Cfg.PNumber: lambda: isinstance(obj, (int, float)) and obj > 0,
        Cfg.NNNumber: lambda: isinstance(obj, (int, float)) and obj >= 0,
        int: lambda: type(obj) == int
    }.get(typ, lambda: isinstance(obj, typ))()


def cfg_isinstance(obj: Any, typ: type | tuple[type]):
    """
    专用于 Cfg 的类型检测

    Args:
        obj (Any): 待检测对象
        typ (Any | tuple[Any]): 类型或类型元组

    Returns:
        bool: 是否为对应类型
    """
    if isinstance(typ, type):
        return cfg_isinstance_single(obj, typ)
    elif isinstance(typ, tuple):
        try:
            for i in typ:
                if cfg_isinstance_single(obj, i):
                    return True
            return False
        except TypeError as e:
            raise ValueError(f"cfg_isinstance arg 2 can't be: {typ}") from e
    else:
        raise ValueError(f"cfg_isinstance arg 2 can't be: {typ}")


def _CfgShowType(typ: Any) -> str:
    """转换类型为中文字符串

    Args:
        typ (Any): 类型

    Returns:
        str: 中文字符串
    """
    if not isinstance(typ, type):
        typ = type(typ)
    return {
        Cfg.PInt: "正整数",
        Cfg.NNInt: "非负整数",
        Cfg.PFloat: "正浮点小数",
        Cfg.NNFloat: "非负浮点小数",
        str: "字符串",
        float: "浮点小数",
        int: "整数",
        dict: "json 对象",
        list: "列表",
        bool: "true/false",
        NoneType: "null",
    }.get(typ, typ.__name__)


class Cfg:
    """配置文件模块"""
    class ConfigError(Exception):
        "配置文件错误"

        def __init__(self, errStr: str, errPos: list | None = None):
            if errPos is None:
                errPos = []
            self.errPos = errPos
            self.args = (errStr,)

    class JsonList:
        "配置文件的列表类型"

        def __init__(self, patt: type | dict | tuple[type | dict, ...], len_limit=-1):
            self.patt = patt
            self.len_limit = len_limit

    class AnyKeyValue:
        "配置文件的任意键名键值对类型"

        def __init__(self, val_type: type | tuple[type] | dict):
            self.type = val_type

    class KeyGroup:
        "配置文件的键组，充当 dict_key 使用"
        def __init__(self, *keys: str):
            self.keys = keys

    class ConfigKeyError(ConfigError):
        "配置 json 的键错误"

    class ConfigValueError(ConfigError):
        "配置 json 的值错误"

    class VersionLowError(ConfigError):
        "配置 json 的版本过低的错误"

    class PInt(int):
        "配置文件的值限制：正整数"

    class NNInt(int):
        "配置文件的值限制：非负整数"

    class PFloat(float):
        "配置文件的值限制：正浮点小数"

    class NNFloat(float):
        "配置文件的值限制：非负浮点小数"

    class PNumber:
        "配置文件的值限制：正数"

    class NNNumber:
        "配置文件的值限制：大于 0 的数"

    class FindNone:
        "找不到值"

    def get_cfg(self, path: str, standard_type: dict):
        "从 path 路径获取 json 文件文本信息，并按照 standard_type 给出的标准形式进行检测。"
        path = path if path.endswith(".json") else path + ".json"
        with open(path, "r", encoding="utf-8") as f:
            try:
                obj = ujson.load(f)
            except ujson.JSONDecodeError as exc:
                raise self.ConfigValueError(
                    "JSON 配置文件格式不正确，请修正或直接删除", None) from exc
        self.check_dict(standard_type, obj)
        return obj

    @staticmethod
    def default_cfg(path: str, default: dict, force: bool = False) -> None:
        """生成默认配置文件

        Args:
            path (str): 路径
            default (dict): 默认配置
            force (bool, optional): 是否强制生成
        """
        path = path if path.endswith(".json") else path + ".json"
        if force or not os.path.isfile(path):
            with open(path, "w", encoding="utf-8") as f:
                ujson.dump(default, f, indent=4, ensure_ascii=False)

    @staticmethod
    def exists(path: str) -> bool:
        """判断文件是否存在

        Args:
            path (str): 路径

        Returns:
            bool: 是否存在
        """
        return os.path.isfile(path if path.endswith(".json") else path + ".json")

    def get_plugin_config_and_version(
        self,
        pluginName: str,
        standardType: Any,
        default: dict,
        default_vers: tuple[int, int, int] | list,
    ) -> tuple[dict[str, Any], tuple[int, int, int]]:
        """获取插件配置文件及版本

        Args:
            pluginName (str): 插件名
            standardType (dict): 标准类型
            default (dict): 默认配置
            default_vers (tuple[int, int, int]): 默认版本

        Returns:
             tuple[dict[str, Any], tuple[int, ...]]: 配置文件及版本
        """
        # 详情见 插件编写指南.md
        assert isinstance(standardType, dict)
        p = "插件配置文件/" + pluginName
        if not self.exists(p) and default:
            defaultCfg = PLUGINCFG_DEFAULT.copy()
            defaultCfg["配置项"] = default
            defaultCfg["配置版本"] = ".".join([str(n) for n in default_vers])
            self.check_auto(standardType, default)
            self.default_cfg(p + ".json", defaultCfg, force=True)
        cfg_stdtyp = PLUGINCFG_STANDARD_TYPE.copy()
        cfg_stdtyp["配置项"] = standardType
        cfgGet = self.get_cfg(p, cfg_stdtyp)
        cfgVers = tuple(int(c) for c in cfgGet["配置版本"].split("."))
        if len(cfgVers) != 3:
            raise ValueError("配置文件出错：版本出错")
        return cfgGet["配置项"], cfgVers

    getPluginConfigAndVersion = get_plugin_config_and_version

    def check_auto(self, standard: type | dict | JsonList | tuple[type | dict, ...], val: Any, fromkey: str = "?"):
        """检查配置文件 (自动类型判断)

        Args:
            standard (type, dict, list): 标准
            val (Any): 值
            fromkey (Any, optional): 键

        Raises:
            ValueError: 未知标准检测类型
            ConfigValueError: 值错误
        """

        if fromkey == Cfg.FindNone:
            raise ValueError("不允许传入 FindNone")
        if isinstance(standard, type):
            if not cfg_isinstance(val, standard):
                if isinstance(val, dict):
                    raise self.ConfigValueError(
                        f'JSON 键"{fromkey}" 对应值的类型不正确：需要 {_CfgShowType(standard)}, 实际上为 json 对象：{ujson.dumps(val, ensure_ascii=False)}'
                    )
                else:
                    raise self.ConfigValueError(
                        f'JSON 键"{fromkey}" 对应值的类型不正确：需要 {_CfgShowType(standard)}, 实际上为 {_CfgShowType(val)}'
                    )
        elif isinstance(standard, Cfg.JsonList):
            self.check_list(standard, val, fromkey)
        elif isinstance(standard, (tuple, list)):
            errs = []
            for single_type in standard:
                try:
                    self.check_auto(single_type, val, fromkey)
                    break
                except Exception as err:
                    errs.append(err)
            else:
                reason = "\n".join(str(err) for err in errs)
                raise self.ConfigValueError(
                    f'JSON 键 对应的键"{fromkey}" 类型不正确，以下为可能的原因：\n{reason}')
        elif isinstance(standard, (dict, Cfg.AnyKeyValue)):
            self.check_dict(standard, val, fromkey)
        else:
            raise ValueError(
                f'JSON 键 "{fromkey}" 自动检测的标准类型传入异常：{standard}'
            )

    def check_dict(self, pattern: dict | AnyKeyValue, jsondict: Any, from_key="?"):
        """
        按照给定的标准配置样式比对传入的配置文件 jsondict, 对不上则引发相应异常

        参数:
            pattern: 标准样式 dict
            jsondict: 待检测的配置文件 dict
        """
        if not isinstance(jsondict, dict):
            raise ValueError(
                f'json 键"{from_key}" 需要 json 对象，而不是 {_CfgShowType(jsondict)}')
        if isinstance(pattern, Cfg.AnyKeyValue):
            for key, val in jsondict.items():
                self.check_auto(pattern.type, val, key)
        else:
            for key, std_val in pattern.items():
                if isinstance(key, self.KeyGroup):
                    for k, v in jsondict.items():
                        if k in key.keys:
                            self.check_auto(std_val, v, k)
                elif isinstance(key, str):
                    val_get = jsondict.get(key, Cfg.FindNone)
                    if val_get == Cfg.FindNone:
                        raise self.ConfigKeyError(f"不存在的 JSON 键：{key}")
                    self.check_auto(std_val, val_get, key)
                else:
                    raise ValueError("Invalid key type: " + key.__class__.__name__)

    def check_list(self, pattern: JsonList, value: Any, fromkey: Any = "?") -> None:
        """检查列表

        Args:
            pattern (list): 标准
            value (Any): 值
            fromkey (Any, optional): 键

        Raises:
            ValueError: 不是合法的标准列表检测样式
            ValueError: 标准检测列表的长度不能为 0
            ConfigValueError: json 键值错误
        """
        if not isinstance(pattern, Cfg.JsonList):
            raise ValueError("不是合法的标准列表检测样式")
        if not isinstance(value, list):
            raise self.ConfigValueError(
                f'JSON 键 "{fromkey}" 需要列表 而不是 {_CfgShowType(value)}'
            )
        if pattern.len_limit != -1:
            if len(value) != pattern.len_limit:
                raise self.ConfigValueError(
                    f'JSON 键 "{fromkey}" 所对应的值列表有误：需要 {pattern.len_limit} 项，实际上为 {len(value)} 项'
                )
        for val in value:
            self.check_auto(pattern.patt, val, fromkey)

    def auto_to_std(self, cfg):
        """
        自动以默认配置文件生成标准配置文件格式.
        注意：不支持固定长度列表以及 Cfg.NeccessaryKey 与 Cfg.Group 的自动转换

        Args:
            cfg: 默认的 CFG 配置文件
        Returns:
            标准 cfg 样式，用于 check_dict
        """
        if isinstance(cfg, dict):
            res = {}
            for k, v in cfg.items():
                if isinstance(v, (dict, list)):
                    res[k] = self.auto_to_std(v)
                elif isinstance(v, (str, int, float, bool)):
                    res[k] = type(v)
            return res
        if isinstance(cfg, list):
            setting_types = []
            for v in cfg:
                if isinstance(v, (dict, list)):
                    t = self.auto_to_std(v)
                else:
                    t = type(v)
                if t not in setting_types:
                    setting_types.append(t)
            if len(setting_types) == 1:
                return Cfg.JsonList(setting_types[0])
            return Cfg.JsonList(tuple(setting_types))
        raise ValueError("auto_to_std() 仅接受 dict 与 list 参数")

    checkDict = check_dict


Config = Cfg()